import hashlib


def compute(query):
    # type (str) -> str
    """
    Computes the query hash via SHA-256.

    Args:
        query: The query

    Returns:
        The query hash
    """
    return hashlib.sha256(query.encode()).hexdigest()



"""
Ported from https://github.com/apollographql/apollo-tooling/blob/master/packages/apollo-graphql/src/operationId.ts
"""
from graphene_tornado.apollo_tooling.transforms import print_with_reduced_whitespace, sort_ast, remove_aliases, hide_literals, \
    drop_unused_definitions


def default_engine_reporting_signature(ast, operation_name):
    """
    The engine reporting signature function consists of removing extra whitespace,
    sorting the AST in a deterministic manner, hiding literals, and removing
    unused definitions.
    """
    return print_with_reduced_whitespace(
      sort_ast(
        remove_aliases(hide_literals(drop_unused_definitions(ast, operation_name)))
      )
    )



"""
Backported from graphql-core-next

https://github.com/graphql-python/graphql-core-next/blob/master/src/graphql/utilities/separate_operations.py
"""

from collections import defaultdict

from graphql.language.ast import Document, OperationDefinition, FragmentDefinition, FragmentSpread
from graphql.language.visitor import Visitor, visit
from typing import Dict, Set

__all__ = ["separate_operations"]


DepGraph = Dict[str, Set[str]]


def separate_operations(document_ast):
    """Separate operations in a given AST document.
    This function accepts a single AST document which may contain many operations and
    fragments and returns a collection of AST documents each of which contains a single
    operation as well the fragment definitions it refers to.
    """

    # Populate metadata and build a dependency graph.
    visitor = SeparateOperations()
    visit(document_ast, visitor)
    operations = visitor.operations
    fragments = visitor.fragments
    positions = visitor.positions
    dep_graph = visitor.dep_graph

    # For each operation, produce a new synthesized AST which includes only what is
    # necessary for completing that operation.
    separated_document_asts = {}
    for operation in operations:
        operation_name = op_name(operation)
        dependencies = set()
        collect_transitive_dependencies(dependencies, dep_graph, operation_name)

        # The list of definition nodes to be included for this operation, sorted to
        # retain the same order as the original document.
        definitions = [operation]
        for name in dependencies:
            definitions.append(fragments[name])
            definitions.sort(key=lambda n: positions.get(n, 0))

        separated_document_asts[operation_name] = Document(definitions=definitions)

    return separated_document_asts


class SeparateOperations(Visitor):

    def __init__(self):
        super(SeparateOperations, self).__init__()
        self.operations = []
        self.fragments = {}
        self.positions = {}
        self.dep_graph = defaultdict(set)
        self.from_name = None
        self.idx = 0

    def enter(self, node, key, parent, path, ancestors):
        if isinstance(node, OperationDefinition):
            self.from_name = op_name(node)
            self.operations.append(node)
            self.positions[node] = self.idx
            self.idx += 1
        elif isinstance(node, FragmentDefinition):
            self.from_name = node.name.value
            self.fragments[self.from_name] = node
            self.positions[node] = self.idx
            self.idx += 1
        elif isinstance(node, FragmentSpread):
            to_name = node.name.value
            self.dep_graph[self.from_name].add(to_name)
        return node


def op_name(operation):
    """Provide the empty string for anonymous operations."""
    return operation.name.value if operation.name else ""


def collect_transitive_dependencies(collected, dep_graph, from_name):
    """Collect transitive dependencies.
    From a dependency graph, collects a list of transitive dependencies by recursing
    through a dependency graph.
    """
    immediate_deps = dep_graph[from_name]
    for to_name in immediate_deps:
        if to_name not in collected:
            collected.add(to_name)
            collect_transitive_dependencies(collected, dep_graph, to_name)



"""
Ported from https://github.com/apollographql/apollo-tooling/blob/master/packages/apollo-graphql/src/transforms.ts
"""
import re

import six

from graphql import print_ast
from graphql.language.ast import Document, IntValue, FloatValue, StringValue, ListValue, ObjectValue, Field, \
    Directive, FragmentDefinition, InlineFragment, FragmentSpread, SelectionSet, OperationDefinition
from graphql.language.visitor import Visitor, visit

from graphene_tornado.apollo_tooling.seperate_operations import separate_operations


def hide_literals(ast):
    # type: (Document) -> Document
    """
    Replace numeric, string, list, and object literals with "empty"
    values. Leaves enums alone (since there's no consistent "zero" enum). This
    can help combine similar queries if you substitute values directly into
    queries rather than use GraphQL variables, and can hide sensitive data in
    your query (say, a hardcoded API key) from Engine servers, but in general
    avoiding those situations is better than working around them.
    """
    visit(ast, _HIDE_LITERALS_VISITOR)
    return ast


def hide_string_and_numeric_literals(ast):
    # type: (Document) -> Document
    """
    In the same spirit as the similarly named `hideLiterals` function, only
    hide string and numeric literals.
    """
    visit(ast, _HIDE_ONLY_STRING_AND_NUMERIC_LITERALS_VISITOR)
    return ast


def drop_unused_definitions(ast, operation_name):
    # type: (Document, str) -> Document
    """
    A GraphQL query may contain multiple named operations, with the operation to
    use specified separately by the client. This transformation drops unused
    operations from the query, as well as any fragment definitions that are not
    referenced.  (In general we recommend that unused definitions are dropped on
    the client before sending to the server to save bandwidth and parsing time.)
    """
    separated = separate_operations(ast).get(operation_name, None)
    if not separated:
        return ast
    return separated


def sort_ast(ast):
    # type: (Document) -> Document
    """
    sortAST sorts most multi-child nodes alphabetically. Using this as part of
    your signature calculation function may make it easier to tell the difference
    between queries that are similar to each other, and if for some reason your
    QraphQL client generates query strings with elements in nondeterministic
    order, it can make sure the queries are treated as identical.
    """
    visit(ast, _SORTING_VISITOR)
    return ast


def remove_aliases(ast):
    # type: (Document) -> Document
    """
    removeAliases gets rid of GraphQL aliases, a feature by which you can tell a
    server to return a field's data under a different name from the field
    name. Maybe this is useful if somebody somewhere inserts random aliases into
    their queries.
    """
    visit(ast, _REMOVE_ALIAS_VISITOR)
    return ast


def print_with_reduced_whitespace(ast):
    # type: (Document) -> str
    """
    Like the graphql-js print function, but deleting whitespace wherever
    feasible. Specifically, all whitespace (outside of string literals) is
    reduced to at most one space, and even that space is removed anywhere except
    for between two alphanumerics.
    """
    visit(ast, _HEX_CONVERSION_VISITOR)
    val = re.sub(r'\s+', ' ',  print_ast(ast))

    val = re.sub(r'([^_a-zA-Z0-9]) ', _replace_with_first_group, val)
    val = re.sub(r' ([^_a-zA-Z0-9])', _replace_with_first_group, val)
    val = re.sub(r'"([a-f0-9]+)"', _from_hex, val)

    return val


def _replace_with_first_group(match):
    return match.group(1)


def _from_hex(match):
    m = match.group(1)
    if six.PY3:
        m = bytes.fromhex(m).decode('utf-8')
    else:
        m = m.decode('hex').encode('utf-8')
    return '"' + m + '"'


def _sorted(items, key):
    if items is not None:
        return sorted(items, key=key)
    return None


class _HideLiteralsVisitor(Visitor):

    def __init__(self, only_string_and_numeric=False):
        super(_HideLiteralsVisitor, self).__init__()
        self._only_string_and_numeric = only_string_and_numeric

    def enter(self, node, key, parent, path, ancestors):
        if isinstance(node, IntValue):
            node.value = 0
        elif isinstance(node, FloatValue):
            node.value = 0
        elif isinstance(node, StringValue):
            node.value = ""
        elif not self._only_string_and_numeric and isinstance(node, ListValue):
            node.values = []
        elif not self._only_string_and_numeric and isinstance(node, ObjectValue):
            node.fields = []
        return node


_HIDE_LITERALS_VISITOR = _HideLiteralsVisitor()
_HIDE_ONLY_STRING_AND_NUMERIC_LITERALS_VISITOR = _HideLiteralsVisitor(only_string_and_numeric=True)


class _RemoveAliasesVisitor(Visitor):

    def enter(self, node, key, parent, path, ancestors):
        if isinstance(node, Field):
           node.alias = None
        return node


_REMOVE_ALIAS_VISITOR = _RemoveAliasesVisitor()


class _HexConversionVisitor(Visitor):

    def enter(self, node, key, parent, path, ancestors):
        if isinstance(node, StringValue) and node.value is not None:
            if six.PY3:
                encoded = node.value.encode('utf-8').hex()
            else:
                encoded = node.value.encode('hex')
            node.value = encoded
        return node


_HEX_CONVERSION_VISITOR = _HexConversionVisitor()


class _SortingVisitor(Visitor):

    def enter(self, node, key, parent, path, ancestors):
        if isinstance(node, Document):
            node.definitions = _sorted(node.definitions, lambda x: (x.__class__.__name__, self._by_name(x)))
        elif isinstance(node, OperationDefinition):
            node.variable_definitions = _sorted(node.variable_definitions, self._by_variable_name)
        elif isinstance(node, SelectionSet):
            node.selections = _sorted(node.selections, lambda x: (x.__class__.__name__, self._by_name(x)))
        elif isinstance(node, Field):
            node.arguments = _sorted(node.arguments, self._by_name)
        elif isinstance(node, FragmentSpread):
            node.directives = _sorted(node.directives, self._by_name)
        elif isinstance(node, InlineFragment):
            node.directives = _sorted(node.directives, self._by_type_definition)
        elif isinstance(node, FragmentDefinition):
            node.directives = _sorted(node.directives, self._by_name)
        elif isinstance(node, Directive):
            node.arguments = _sorted(node.arguments, self._by_name)
        return node

    def _by_name(self, node):
        if isinstance(node, InlineFragment):
            return self._by_type_definition(node)
        elif node.name is not None:
            return node.name.value
        return None

    def _by_type_definition(self, node):
        if node.type_condition is not None and node.type_condition.name:
            return node.type_condition.name.value
        return None

    def _by_variable_name(self, node):
        if node.variable is not None and node.variable.name is not None:
            return node.variable.name.value
        return None


_SORTING_VISITOR = _SortingVisitor()



from graphql import parse

from graphene_tornado.apollo_tooling.operation_id import default_engine_reporting_signature


def test_basic_signature():
    operation = ""
    doc = parse("""
        {
          user {
            name
          }
        }    
    """)
    assert "{user{name}}" == default_engine_reporting_signature(doc, operation)


def test_basic_signature_with_query():
    operation = ""
    doc = parse("""
        query {
          user {
            name
          }
        }    
    """)
    assert "{user{name}}" == default_engine_reporting_signature(doc, operation)


def test_basic_signature_with_operation_name():
    operation = "OpName"
    doc = parse("""
        query OpName {
          user {
            name
          }
        }
    """)
    assert "query OpName{user{name}}" == default_engine_reporting_signature(doc, operation)


def test_basic_signature_with_fragment():
    operation = ""
    doc = parse("""
        {
          user {
            name
            ...Bar
          }
        }
        fragment Bar on User {
          asd
        }
        fragment Baz on User {
          jkl
        }
    """)
    assert "fragment Bar on User{asd}{user{name...Bar}}" == default_engine_reporting_signature(doc, operation)


def test_basic_signature_with_fragments_in_various_order():
    operation = ""
    doc = parse("""
        fragment Bar on User {
          asd
        }
        {
          user {
            name
            ...Bar
          }
        }
        fragment Baz on User {
          jkl
        }
    """)
    assert "fragment Bar on User{asd}{user{name...Bar}}" == default_engine_reporting_signature(doc, operation)


def test_basic_signature_with_full_test():
    operation = "Foo"
    doc = parse("""
        query Foo($b: Int, $a: Boolean) {
          user(name: "hello", age: 5) {
            ...Bar
            ... on User {
              hello
              bee
            }
            tz
            aliased: name
          }
        }
        fragment Baz on User {
          asd
        }
        fragment Bar on User {
          age @skip(if: $a)
          ...Nested
        }
        fragment Nested on User {
          blah
        }
    """)
    assert "fragment Bar on User{age@skip(if:$a)...Nested}fragment Nested on User{blah}query Foo($a:Boolean,$b:Int)" \
           "{user(age:0,name:\"\"){name tz...Bar...on User{bee hello}}}"\
           == default_engine_reporting_signature(doc, operation)


def test_basic_signature_with_various_argument_types():
    operation = "OpName"
    doc = parse("""
        query OpName($c: Int!, $a: [[Boolean!]!], $b: EnumType) {
          user {
            name(apple: $a, cat: $c, bag: $b)
          }
        }
    """)
    assert "query OpName($a:[[Boolean!]!],$b:EnumType,$c:Int!){user{name(apple:$a,bag:$b,cat:$c)}}" \
           == default_engine_reporting_signature(doc, operation)


def test_basic_signature_with_inline_types():
    operation = "OpName"
    doc = parse("""
        query OpName {
          user {
            name(apple: [[10]], cat: ENUM_VALUE, bag: { input: "value" })
          }
        }
    """)
    assert "query OpName{user{name(apple:[],bag:{},cat:ENUM_VALUE)}}" == \
           default_engine_reporting_signature(doc, operation)






import collections

from graphene_tornado.apollo_tooling.operation_id import default_engine_reporting_signature
from graphene_tornado.request_context import SIGNATURE


class _LRUCache:
    def __init__(self, capacity):
        self.capacity = capacity
        self.cache = collections.OrderedDict()

    def get(self, key, default=None):
        try:
            value = self.cache.pop(key)
            self.cache[key] = value
            return value
        except KeyError:
            return default

    def set(self, key, value):
        try:
            self.cache.pop(key)
        except KeyError:
            if len(self.cache) >= self.capacity:
                self.cache.popitem(last=False)
        self.cache[key] = value


_SIGNATURE_CACHE = _LRUCache(10000)


def get_signature(request_context, operation_name, document, query_string):
    """
    Args:
        request_context: The request context
        operation_name: The operation name
        document: The document
        query_string: The query string
    Returns:
        The signature for the query
    """
    signature = request_context.get(SIGNATURE, None)
    if signature is None:
        signature = _SIGNATURE_CACHE.get(query_string)
    if signature is None:
        if document:
            calculate_signature = default_engine_reporting_signature
            signature = calculate_signature(document.document_ast, operation_name)
        elif query_string:
            signature = query_string
        request_context[SIGNATURE] = signature
    return signature





"""
ExtensionStack is an adapter for GraphQLExtension that helps invoke a list of GraphQLExtension objects at runtime.
"""
import inspect
from typing import List, Callable, Union

from graphene_tornado.graphql_extension import GraphQLExtension


def instantiate_extensions(extensions):
    for extension in extensions:
        if inspect.isclass(extension) or inspect.isfunction(extension):
            yield extension()
            continue
        yield extension


class GraphQLExtensionStack(GraphQLExtension):

    def __init__(self,
                 extensions=None  # type: List[Union[Callable[[], GraphQLExtension], GraphQLExtension]]
                 ):
        self.extensions = list(instantiate_extensions(extensions))  # type: List[GraphQLExtension]

    async def request_started(self, request, query_string, parsed_query, operation_name, variables, context, request_context):
        on_end = await self._handle_did_start('request_started', request, query_string, parsed_query, operation_name,
                                              variables, context, request_context)
        return on_end

    async def parsing_started(self, query_string):
        on_end = await self._handle_did_start('parsing_started', query_string)
        return on_end

    async def validation_started(self):
        on_end = await self._handle_did_start('validation_started')
        return on_end

    async def execution_started(self, schema, document, root, context, variables, operation_name, request_context):
        on_end = await self._handle_did_start('execution_started', schema, document, root, context,
                                              variables, operation_name, request_context)
        return on_end

    async def will_resolve_field(self, root, info, **args):
        ext = self.extensions[:]
        ext.reverse()

        for extension in self.extensions:
            on_end = await extension.will_resolve_field(root, info, **args)
            await on_end()

        async def on_end(error=None, result=None):
            return (error, result)

        return on_end

    async def will_send_response(self, response, context):
        ref = [response, context]
        ext = self.extensions[:]
        ext.reverse()

        for handler in ext:
            result = await handler.will_send_response(ref[0], ref[1])
            if result:
                ref = [result, context]
        return ref

    async def _handle_did_start(self, method, *args):
        end_handlers = []
        for extension in self.extensions:
            invoker = getattr(extension, method)
            end_handler = await invoker(*args)
            if end_handler:
                end_handlers.append(end_handler)

        async def end(errors=None):
            errors = errors or []
            end_handlers.reverse()
            for handler in end_handlers:
                if handler:
                    await handler(errors)

        return end



"""
GraphQLExtension is analogous to the server extensions that are provided
by Apollo Server: https://github.com/apollographql/apollo-server/tree/master/packages/graphql-extensions

Extensions are also middleware but have additional hooks.
"""
from __future__ import absolute_import, print_function

from abc import ABCMeta, abstractmethod
from typing import NewType, List, Callable, Optional, Dict, Any

from graphql import GraphQLSchema
from graphql.language.ast import Document
from tornado.httputil import HTTPServerRequest

EndHandler = NewType('EndHandler', Optional[List[Callable[[List[Exception]], None]]])


class GraphQLExtension:

    __metaclass__ = ABCMeta

    @abstractmethod
    def request_started(self,
                        request,         # type: HTTPServerRequest
                        query_string,    # type: Optional[str],
                        parsed_query,    # type: Optional[Document]
                        operation_name,  # type: Optional[str]
                        variables,       # type: Optional[dict[str, Any]]
                        context,         # type: Any
                        request_context  # type: Any
                        ):
        # type: (...) -> EndHandler
        pass

    @abstractmethod
    def parsing_started(self, query_string):  # type: (str) -> EndHandler
        pass

    @abstractmethod
    def validation_started(self):
        # type: () -> EndHandler
        pass

    @abstractmethod
    def execution_started(self,
                          schema,  # type: GraphQLSchema
                          document,  # type: Document
                          root,  # type: Any
                          context,  # type: Optional[Any]
                          variables,  # type: Optional[Any]
                          operation_name,  # type: Optional[str]
                          request_context  # type: Dict[Any, Any]
                          ):
        # type: (...) -> EndHandler
        pass

    @abstractmethod
    def will_resolve_field(self, root, info, **args):
        # type: (...) -> EndHandler
        pass

    @abstractmethod
    def will_send_response(self,
                           response,  # type: Any
                           context,   # type: Any
                           ):
        # type: (...) -> EndHandler
        pass

    def as_middleware(self):
        # type: () -> Callable
        """
        Adapter for using the stack as middleware so that the will_resolve_field function
        is invoked like normal middleware

        Returns:
            An adapter function that acts as middleware
        """
        async def middleware(next, root, info, **args):
            end_resolve = await self.will_resolve_field(root, info, **args)
            res = None
            errors = []
            try:
                res = next(root, info, **args)
                return res
            finally:
                await end_resolve(errors, res)
        return middleware



from __future__ import absolute_import, division, print_function

from graphql.type.definition import GraphQLArgument, GraphQLField, GraphQLNonNull, GraphQLObjectType
from graphql.type.scalars import GraphQLString
from graphql.type.schema import GraphQLSchema
from tornado.gen import coroutine, Return


async def resolve_raises(*_):
    raise Exception("Throws!")


async def resolve1(obj, args, context, info):
    return context.args.get('q')


async def resolve2(obj, args, context, info):
    return context


async def resolve3(obj, args, context, info):
    return 'Hello %s' % (args.get('who') or 'World')


QueryRootType = GraphQLObjectType(
    name='QueryRoot',
    fields={
        'thrower': GraphQLField(GraphQLNonNull(GraphQLString), resolver=resolve_raises),
        'request': GraphQLField(GraphQLNonNull(GraphQLString),
                                resolver=resolve1),
        'context': GraphQLField(GraphQLNonNull(GraphQLString),
                                resolver=resolve2),
        'test': GraphQLField(
            type=GraphQLString,
            args={
                'who': GraphQLArgument(GraphQLString)
            },
            resolver=resolve3
        )
    }
)

MutationRootType = GraphQLObjectType(
    name='MutationRoot',
    fields={
        'writeTest': GraphQLField(
            type=QueryRootType,
            resolver=lambda *_: QueryRootType
        )
    }
)

Schema = GraphQLSchema(QueryRootType, MutationRootType)



import pytest
import tornado
from graphql import build_ast_schema, parse

from graphene_tornado.ext.apollo_engine_reporting.engine_agent import EngineReportingOptions, EngineReportingAgent
from graphene_tornado.ext.apollo_engine_reporting.engine_extension import EngineReportingExtension
from graphene_tornado.ext.apollo_engine_reporting.tests.schema import SCHEMA_STRING
from graphene_tornado.ext.apollo_engine_reporting.tests.test_engine_extension import QUERY
from graphene_tornado.ext.apollo_engine_reporting.schema_utils import generate_schema_hash
from graphene_tornado.tests.http_helper import HttpHelper
from graphene_tornado.tests.test_graphql import response_json, url_string, GRAPHQL_HEADER
from graphene_tornado.tornado_graphql_handler import TornadoGraphQLHandler

SCHEMA = build_ast_schema(parse(SCHEMA_STRING))


engine_options = EngineReportingOptions(api_key='test')
agent = EngineReportingAgent(engine_options, generate_schema_hash(SCHEMA))


class ExampleEngineReportingApplication(tornado.web.Application):

    def __init__(self):
        engine_extension = EngineReportingExtension(engine_options, agent.add_trace)
        handlers = [
            (r'/graphql', TornadoGraphQLHandler, dict(graphiql=True, schema=SCHEMA, extensions=[engine_extension])),
            (r'/graphql/batch', TornadoGraphQLHandler, dict(graphiql=True, schema=SCHEMA, batch=True)),
        ]
        tornado.web.Application.__init__(self, handlers)


@pytest.fixture
def app():
    return ExampleEngineReportingApplication()


@pytest.fixture
def http_helper(http_client, base_url):
    return HttpHelper(http_client, base_url)


@pytest.mark.gen_test()
@pytest.mark.skip()
def test_can_send_report_to_engine(http_helper):
    response = yield http_helper.get(url_string(query=QUERY), headers=GRAPHQL_HEADER)
    assert response.code == 200
    assert 'data' in response_json(response)



import tornado
from tornado.ioloop import IOLoop

from graphene_tornado.schema import schema
from graphene_tornado.ext.apollo_engine_reporting import EngineReportingOptions, EngineReportingAgent
from graphene_tornado.ext.apollo_engine_reporting.engine_extension import EngineReportingExtension
from graphene_tornado.ext.apollo_engine_reporting.schema_utils import generate_schema_hash
from graphene_tornado.tornado_graphql_handler import TornadoGraphQLHandler

engine_options = EngineReportingOptions()
agent = EngineReportingAgent(engine_options, generate_schema_hash(schema))


class ExampleEngineReportingApplication(tornado.web.Application):

    def __init__(self):
        engine_extension = lambda: EngineReportingExtension(engine_options, agent.add_trace)
        handlers = [
            (r'/graphql', TornadoGraphQLHandler, dict(graphiql=True, schema=schema, extensions=[engine_extension])),
            (r'/graphql/batch', TornadoGraphQLHandler, dict(graphiql=True, schema=schema, batch=True,
                                                            extensions=[engine_extension])),
            (r'/graphql/graphiql', TornadoGraphQLHandler, dict(graphiql=True, schema=schema,
                                                               extensions=[engine_extension]))
        ]
        tornado.web.Application.__init__(self, handlers)


if __name__ == '__main__':
    app = ExampleEngineReportingApplication()
    app.listen(5000)
    IOLoop.instance().start()



import re

import pytest
import six
import tornado
from google.protobuf.json_format import MessageToJson
from graphql import parse

from graphene_tornado.ext.apollo_engine_reporting.engine_agent import EngineReportingOptions
from graphene_tornado.ext.apollo_engine_reporting.engine_extension import EngineReportingExtension
from graphene_tornado.ext.apollo_engine_reporting.tests.schema import schema
from graphene_tornado.tests.http_helper import HttpHelper
from graphene_tornado.tests.test_graphql import GRAPHQL_HEADER, url_string, response_json
from graphene_tornado.tornado_graphql_handler import TornadoGraphQLHandler

expected = """{
  "durationNs": "-1", 
  "endTime": "2019-09-21T19:37:09.908919Z", 
  "http": {
    "method": "GET"
  }, 
  "root": {
    "child": [
      {
        "child": [
          {
            "endTime": "-1", 
            "parentType": "User", 
            "responseName": "name", 
            "startTime": "-1", 
            "type": "String"
          }, 
          {
            "child": [
              {
                "child": [
                  {
                    "endTime": "-1", 
                    "parentType": "Post", 
                    "responseName": "id", 
                    "startTime": "-1", 
                    "type": "Int"
                  }
                ], 
                "index": 0
              }, 
              {
                "child": [
                  {
                    "endTime": "-1", 
                    "parentType": "Post", 
                    "responseName": "id", 
                    "startTime": "-1", 
                    "type": "Int"
                  }
                ], 
                "index": 1
              }
            ], 
            "endTime": "-1", 
            "parentType": "User", 
            "responseName": "posts", 
            "startTime": "-1", 
            "type": "[Post]"
          }
        ], 
        "endTime": "-1", 
        "parentType": "Query", 
        "responseName": "author", 
        "startTime": "-1", 
        "type": "User"
      }, 
      {
        "endTime": "-1", 
        "parentType": "Query", 
        "responseName": "aBoolean", 
        "startTime": "-1", 
        "type": "Boolean"
      }
    ], 
    "endTime": "-1", 
    "startTime": "-1"
  }, 
  "startTime": "2019-09-21T19:37:09.908919Z"
}"""


QUERY = """
    query {
      author(id: 5) {
        name
        posts(limit: 2) {
          id
        }
      }
      aBoolean
    }
"""


engine_options = EngineReportingOptions(api_key='test')

traces = []


async def add_trace(operation_name, document_ast, query_string, trace):
    traces.append((operation_name, document_ast, query_string, trace))


class ExampleEngineReportingApplication(tornado.web.Application):

    def __init__(self):
        engine_extension = lambda: EngineReportingExtension(engine_options, add_trace)
        handlers = [
            (r'/graphql', TornadoGraphQLHandler, dict(graphiql=True, schema=schema, extensions=[engine_extension])),
            (r'/graphql/batch', TornadoGraphQLHandler, dict(graphiql=True, schema=schema, batch=True)),
        ]
        tornado.web.Application.__init__(self, handlers)


@pytest.fixture
def app():
    return ExampleEngineReportingApplication()


@pytest.fixture
def http_helper(http_client, base_url):
    return HttpHelper(http_client, base_url)


@pytest.mark.gen_test()
def test_can_send_report_to_engine(http_helper):
    response = yield http_helper.get(url_string(query=QUERY), headers=GRAPHQL_HEADER)
    assert response.code == 200
    assert 'data' in response_json(response)

    assert len(traces) == 1
    operation_name, document_ast, query_string, trace = traces[0]

    assert QUERY == query_string
    assert parse(QUERY) == document_ast
    assert '' == operation_name

    trace_json = MessageToJson(trace, sort_keys=True)
    trace_json = re.sub(r'"([0-9]+)"', '"-1"', trace_json)
    trace_json = re.sub(r'"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d+Z"',
                        '"2019-09-21T19:37:09.908919Z"',
                        trace_json)

    e = expected
    if six.PY3:
        e = re.sub(r'\s+\n', '\n', expected)
    assert e == trace_json



from __future__ import absolute_import

import gzip

import pytest
import six
import tornado
from graphql import parse
from six import StringIO, BytesIO

from graphene_tornado.apollo_tooling.operation_id import default_engine_reporting_signature
from graphene_tornado.ext.apollo_engine_reporting.engine_agent import EngineReportingAgent, EngineReportingOptions, \
    SERVICE_HEADER_DEFAULTS
from graphene_tornado.ext.apollo_engine_reporting.engine_extension import EngineReportingExtension
from graphene_tornado.ext.apollo_engine_reporting.reports_pb2 import FullTracesReport
from graphene_tornado.ext.apollo_engine_reporting.tests.schema import schema
from graphene_tornado.ext.apollo_engine_reporting.tests.test_engine_extension import QUERY
from graphene_tornado.tests.http_helper import HttpHelper
from graphene_tornado.tests.test_graphql import response_json, url_string, GRAPHQL_HEADER
from graphene_tornado.tornado_graphql_handler import TornadoGraphQLHandler


class RecordingEngineReportingAgent(EngineReportingAgent):

    def __init__(self, options, schema_hash):
        EngineReportingAgent.__init__(self, options, schema_hash)
        self.data = []

    async def post_data(self, data):
        self.data.append(data)

    def reset(self):
        self.data = []


engine_options = EngineReportingOptions(api_key='test')
agent = RecordingEngineReportingAgent(engine_options, "hash")


class ExampleEngineReportingApplication(tornado.web.Application):

    def __init__(self):
        engine_extension = lambda: EngineReportingExtension(engine_options, agent.add_trace)
        handlers = [
            (r'/graphql', TornadoGraphQLHandler, dict(graphiql=True, schema=schema, extensions=[engine_extension])),
            (r'/graphql/batch', TornadoGraphQLHandler, dict(graphiql=True, schema=schema, batch=True)),
        ]
        tornado.web.Application.__init__(self, handlers)


@pytest.fixture
def app():
    return ExampleEngineReportingApplication()


@pytest.fixture
def http_helper(http_client, base_url):
    return HttpHelper(http_client, base_url)


@pytest.fixture(scope="function", autouse=True)
def reset():
    yield
    agent.reset()


@pytest.mark.gen_test
def test_can_send_report_to_engine(http_helper):
    response = yield http_helper.get(url_string(query=QUERY), headers=GRAPHQL_HEADER)

    report = _deserialize(agent.data[0])

    assert report.header.hostname == SERVICE_HEADER_DEFAULTS.get('hostname')
    assert report.header.agent_version == SERVICE_HEADER_DEFAULTS.get('agentVersion')
    assert report.header.runtime_version == SERVICE_HEADER_DEFAULTS.get('runtimeVersion')
    assert report.header.uname == SERVICE_HEADER_DEFAULTS.get('uname')

    assert len(report.traces_per_query) == 1
    key = next(iter(report.traces_per_query))
    query_key = '# -\n' + default_engine_reporting_signature(parse(QUERY), '')
    assert query_key == key
    assert report.traces_per_query[query_key].trace.pop()

    assert response.code == 200
    assert 'data' in response_json(response)


def _deserialize(message):
    fileobj = BytesIO(message) if six.PY3 else StringIO(message)
    content = gzip.GzipFile(fileobj=fileobj).read()
    return FullTracesReport.FromString(content)



from graphene import Field, List, Int, Boolean, String, ObjectType, Schema

SCHEMA_STRING = """
  schema {
      query: Query
  }
  
  type User {
    id: Int
    name: String
    posts(limit: Int): [Post]
  }
  type Post {
    id: Int
    title: String
    views: Int
    author: User
  }
  type Query {
    aString: String
    aBoolean: Boolean
    anInt: Int
    author(id: Int): User
    topPosts(limit: Int): [Post]
  }
"""


class User(ObjectType):

    id = Int()
    name = String()
    posts = Field(lambda: List(Post), limit=Int())

    async def get_node(self, info, id):
        return User(id=id)

    def resolve_posts(self, info, **args):
        return [Post(id=1), Post(id=2)]


class Post(ObjectType):
    """A ship in the Star Wars saga"""

    id = Int()
    title = String()
    views = Int()
    author = Field(User)

    async def get_node(self, info, **args):
        return Post(id=1)


class Query(ObjectType):
    a_string = Field(String, name='aString')
    a_boolean = Field(Boolean, name='aBoolean')
    an_int = Field(Int, name='anInt')
    author = Field(User, id=Int())
    top_posts = Field(List(Post), limit=Int())

    async def resolve_author(self, info, **args):
        return User()

    async def resolve_posts(self, info, **args):
        return [Post(id=1), Post(id=2)]


schema = Schema(query=Query)




import json
from functools import partial

import pytest
import tornado

from graphene_tornado.graphql_extension import GraphQLExtension
from graphene_tornado.schema import schema
from graphene_tornado.tests.http_helper import HttpHelper
from graphene_tornado.tests.test_graphql import url_string, GRAPHQL_HEADER, response_json
from graphene_tornado.tornado_graphql_handler import TornadoGraphQLHandler

STARTED = []
ENDED = []


async def _track_closed(name, errors=None):
    ENDED.append(name)


class TrackingExtension(GraphQLExtension):

    async def request_started(self, request, query_string, parsed_query, operation_name, variables, context, request_context):
        phase = 'request'
        STARTED.append(phase)
        return partial(_track_closed, phase)

    async def parsing_started(self, query_string):
        phase = 'parsing'
        STARTED.append(phase)
        return partial(_track_closed, phase)

    async def validation_started(self):
        phase = 'validation'
        STARTED.append(phase)
        return partial(_track_closed, phase)

    async def execution_started(self, schema, document, root, context, variables, operation_name, request_context):
        phase = 'execution'
        STARTED.append(phase)
        return partial(_track_closed, phase)

    async def will_resolve_field(self, root, info, **args):
        phase = 'resolve_field'
        STARTED.append(phase)
        return partial(_track_closed, phase)

    async def will_send_response(self, response, context):
        phase = 'response'
        STARTED.append(phase)


class ExampleExtensionsApplication(tornado.web.Application):

    def __init__(self):
        handlers = [
            (r'/graphql', TornadoGraphQLHandler, dict(graphiql=True, schema=schema, extensions=[TrackingExtension()])),
            (r'/graphql/batch', TornadoGraphQLHandler, dict(graphiql=True, schema=schema, batch=True)),
            (r'/graphql/graphiql', TornadoGraphQLHandler, dict(graphiql=True, schema=schema))
        ]
        tornado.web.Application.__init__(self, handlers)


@pytest.fixture
def app():
    return ExampleExtensionsApplication()


@pytest.fixture
def http_helper(http_client, base_url):
    return HttpHelper(http_client, base_url)


@pytest.mark.gen_test
def test_extensions_are_called_in_order(http_helper):
    response = yield http_helper.get(url_string(
        query='query helloWho($who: String){ test(who: $who) }',
        variables=json.dumps({'who': "Dolly"})
    ), headers=GRAPHQL_HEADER)

    assert response.code == 200
    assert response_json(response) == {
        'data': {'test': "Hello Dolly"}
    }

    assert ['request', 'parsing', 'validation', 'execution', 'resolve_field', 'response'] == STARTED
    assert ['parsing', 'validation', 'execution', 'resolve_field', 'request'] == ENDED




from graphql import parse

from graphene_tornado.apollo_tooling.transforms import print_with_reduced_whitespace, hide_literals


def test_print_with_reduced_whitespace():
    doc = parse('''
        query Foo($a: Int) {
          user(
            name: "   tab->	yay"
          ) {
            name
          }
        }        
    ''')
    transformed = print_with_reduced_whitespace(doc)
    assert 'query Foo($a:Int){user(name:"   tab->\tyay"){name}}' == transformed


def test_hide_literals():
    doc = parse('''
        query Foo($b: Int, $a: Boolean) {
          user(name: "hello", age: 5) {
            ...Bar
            ... on User {
              hello
              bee
            }
            tz
            aliased: name
          }
        }
        fragment Bar on User {
          age @skip(if: $a)
          ...Nested
        }
        fragment Nested on User {
          blah
        }    
    ''')
    transformed = print_with_reduced_whitespace(hide_literals(doc))
    assert 'query Foo($b:Int,$a:Boolean){user(name:"",age:0){...Bar...on User{hello bee}tz aliased:name}}fragment ' \
           'Bar on User{age@skip(if:$a)...Nested}fragment Nested on User{blah}' == transformed







from __future__ import absolute_import

import json

from opencensus.trace import execution_context
from tornado.gen import Return, coroutine

from graphene_tornado.apollo_tooling.query_hash import compute
from graphene_tornado.ext.extension_helpers import get_signature
from graphene_tornado.extension_stack import GraphQLExtension
from graphene_tornado.request_context import SIGNATURE_HASH_KEY


class OpenCensusExtension(GraphQLExtension):

    def __init__(self):
        self.operation_name = None
        self.query_string = None
        self.document = None

    async def request_started(self, request, query_string, parsed_query, operation_name, variables, context, request_context):
        self.query_string = query_string
        self.document = parsed_query

        tracer = execution_context.get_opencensus_tracer()
        tracer.start_span('gql')

        async def on_request_ended(errors):
            op_name = self.operation_name or ''
            document = request_context.get('document', None)
            signature = get_signature(request_context, operation_name, document, query_string)

            tracer = execution_context.get_opencensus_tracer()
            if SIGNATURE_HASH_KEY not in request_context:
                request_context[SIGNATURE_HASH_KEY] = compute(signature)
            tracer.current_span().name = 'gql[{}]'.format(request_context[SIGNATURE_HASH_KEY][0:12])

            tracer.add_attribute_to_current_span('gql_operation_name', op_name)
            tracer.add_attribute_to_current_span('signature', signature)
            tracer.end_span()

        return on_request_ended

    async def parsing_started(self, query_string):
        pass

    async def validation_started(self):
        pass

    async def execution_started(self, schema, document, root, context, variables, operation_name, request_context):
        if operation_name:
            self.operation_name = operation_name
        request_context['document'] = document

    async def will_resolve_field(self, root, info, **args):
        if not self.operation_name:
            self.operation_name = '' if not info.operation.name else info.operation.name.value

        tracer = execution_context.get_opencensus_tracer()

        # If we wanted to be fancy, we could build up a tree like the ApolloEngineExtension does so that the
        # whole tree appears as nested spans. However, this is a bit tricky to do with the current OpenCensus
        # API because when you request a span, a bunch of context variables are set. This keeps it simple for now.
        tracer.start_span('.'.join(str(x) for x in info.path))

        async def on_end(errors=None, result=None):
            tracer.end_span()

        return on_end

    async def will_send_response(self, response, context):
        if hasattr(response, 'errors'):
            errors = response.errors
            for error in errors:
                error_info = {
                    'message': str(error),
                    'json': error
                }

                tracer = execution_context.get_opencensus_tracer()
                tracer.add_attribute_to_current_span('gql_error', json.dumps(error_info))



from __future__ import absolute_import, print_function

import gzip
import logging
import os
import socket
import sys
from typing import NamedTuple, Optional, Callable

import six
from google.protobuf.json_format import MessageToJson
from google.protobuf.message import Message
from six import StringIO, BytesIO
from tornado.httpclient import AsyncHTTPClient
from tornado_retry_client import RetryClient

from graphene_tornado.apollo_tooling.operation_id import default_engine_reporting_signature
from .reports_pb2 import ReportHeader, FullTracesReport

LOGGER = logging.getLogger(__name__)

SERVICE_HEADER_DEFAULTS = {
  'hostname': socket.gethostname(),
  'agentVersion': 'apollo-engine-reporting@${require(\'../package.json\').version}',
  'runtimeVersion': 'python ' + '.'.join(map(str, sys.version_info[0:3])),
  'uname': ' '.join(os.uname()),
}

GenerateClientInfo = NamedTuple('GenerateClientInfo', [
    ('client_name', Optional[str]),
    ('client_version', Optional[str]),
    ('client_reference_id', Optional[str]),
])

EngineReportingOptions = NamedTuple('EngineReportingOptions', [
    ('api_key', Optional[str]),
    ('calculate_signature', Optional[Callable]),
    # ('report_interval_ms', Optional[int]),
    # ('max_uncompressed_report_size', Optional[int]),
    ('endpoint_url', Optional[str]),
    ('debug_print_reports', Optional[bool]),
    ('request_agent', Optional[bool]),
    # ('max_attempts', Optional[int]),
    # ('minimum_retry_delay_ms', Optional[int]),
    ('report_error_function', Optional[Callable]),
    # ('private_variables', Optional[List[str]]),
    # ('private_headers', Optional[List[str]]),
    # ('handle_signals', Optional[bool]),
    # ('send_reports_immediately', Optional[bool]),
    ('mask_errors_details', Optional[bool]),
    ('schema_tag', Optional[str]),
    ('generate_client_info', Optional[GenerateClientInfo])
])
EngineReportingOptions.__new__.__defaults__ = (None,) * len(EngineReportingOptions._fields)


def _serialize(message):
    # type: (Message) -> bytes
    out = BytesIO() if six.PY3 else StringIO()
    with gzip.GzipFile(fileobj=out, mode="w") as f:
        f.write(message.SerializeToString())
    return out.getvalue()


def _get_trace_signature(operation_name, document_ast, query_string):
    if not document_ast:
        return query_string
    else:
        return default_engine_reporting_signature(document_ast, operation_name)


class EngineReportingAgent:

    def __init__(self, options, schema_hash):  # type: (EngineReportingOptions, str) -> None
        self.options = options
        self.api_key = options.api_key or os.getenv('ENGINE_API_KEY', None)

        if not self.api_key:
            raise ValueError('To use EngineReportingAgent, you must specify an API key via the api_key option or the ' +
                             'ENGINE_API_KEY environment variable.')

        self.endpoint_url = self.options.endpoint_url or 'https://engine-report.apollodata.com/api/ingress/traces'
        self.request_headers = {
            'user-agent': 'apollo-engine-reporting',
            'x-api-key': self.api_key,
            'content-encoding': 'gzip',
        }

        self._stopped = False

        self.report_header = ReportHeader()
        self.report_header.hostname = SERVICE_HEADER_DEFAULTS['hostname']
        self.report_header.agent_version = SERVICE_HEADER_DEFAULTS['agentVersion']
        self.report_header.runtime_version = SERVICE_HEADER_DEFAULTS['runtimeVersion']
        self.report_header.uname = SERVICE_HEADER_DEFAULTS['uname']
        self.report_header.schema_hash = schema_hash
        self.report_header.schema_tag = options.schema_tag or os.getenv('ENGINE_SCHEMA_TAG', None) or ''

        self.report = FullTracesReport(header=self.report_header)
        self.report_size = 0

    def _options(self) -> EngineReportingOptions:
        return self.options

    async def add_trace(self, operation_name, document_ast, query_string, trace):
        operation_name = operation_name or '-'

        if self._stopped:
            return

        signature = _get_trace_signature(operation_name, document_ast, query_string)
        stats_report_key = "# " + operation_name + '\n' + signature
        traces_per_query = self.report.traces_per_query.get(stats_report_key, None)
        if not traces_per_query:
            traces_per_query = self.report.traces_per_query[stats_report_key]
        traces_per_query.trace.extend([trace])

        await self.send_report_and_report_errors()

    async def send_report(self):
        report = self.report
        self.reset_report()

        if len(report.traces_per_query) == 0:
            return

        if self.options.debug_print_reports:
            LOGGER.info('Engine sending report: ' + MessageToJson(report))

        await self.post_data(_serialize(report))

    async def post_data(self, data):
        headers = {
            'Content-Length': len(data)
        }
        headers.update(self.request_headers)

        http_client = AsyncHTTPClient()
        retry_client = RetryClient(
            http_client=http_client,
            retry_attempts=3,
            retry_start_timeout=0.5,
            retry_max_timeout=10,
            retry_factor=2,
        )

        try:
            response = await retry_client.fetch(self.endpoint_url, method='POST', headers=headers, body=data,
                                                raise_error=False)

        finally:
            http_client.close()

        if 500 <= response.code < 600:
            raise ValueError(response.code + ': ' + response.body)

        if response.code < 200 or response.code >= 300:
            raise ValueError('Error sending report to Engine servers (HTTP status {}): {}'
                             .format(response.code, response.body))

        if self.options.debug_print_reports:
            LOGGER.info('Engine report: status ' + response.code)

    def stop(self):
        self._stopped = True

    async def send_report_and_report_errors(self):
        try:
            await self.send_report()
        except:
            exception = sys.exc_info()[1]
            if self.options.report_error_function:
                self.options.report_error_function(exception)
            else:
                LOGGER.exception("Error sending reports to Apollo Engine")

    def reset_report(self):
        self.report = FullTracesReport(header=self.report_header)
        self.report_size = 0  # type: int





from __future__ import absolute_import, print_function

import json
import time
from numbers import Number
from typing import Callable, NamedTuple

from google.protobuf.timestamp_pb2 import Timestamp
from tornado.httputil import HTTPServerRequest

from graphene_tornado.ext.apollo_engine_reporting.engine_agent import EngineReportingOptions
from graphene_tornado.ext.apollo_engine_reporting.reports_pb2 import Trace
from graphene_tornado.graphql_extension import GraphQLExtension

CLIENT_NAME_HEADER = 'apollographql-client-name'
CLIENT_REFERENCE_HEADER_KEY = 'apollographql-client-reference-id'
CLIENT_VERSION_HEADER_KEY = 'apollographql-client-version'


ClientInfo = NamedTuple('EngineReportingOptions', [
    ('client_name', str),
    ('client_reference_id', str),
    ('client_version', str)
])


def generate_client_info(request):
    # type: (HTTPServerRequest) -> ClientInfo
    return ClientInfo(
        request.headers.get(CLIENT_NAME_HEADER, ''),
        request.headers.get(CLIENT_REFERENCE_HEADER_KEY, ''),
        request.headers.get(CLIENT_VERSION_HEADER_KEY, ''),
    )


def response_path_as_string(path):
    if not path:
        return ''
    return '.'.join((str(x) for x in path))


def now_ns():
    return time.time_ns()


class EngineReportingExtension(GraphQLExtension):

    def __init__(self, options, add_trace):  # type: (EngineReportingOptions, Callable) -> None
        if add_trace is None:
            raise ValueError('add_trace must be defined')

        self.add_trace = add_trace
        self.operation_name = None

        self.options = options  # maskErrorDetails = False
        self.start_time = now_ns()

        root = Trace.Node()
        root.start_time = self.start_time
        self.trace = Trace(root=root)
        self.nodes = {response_path_as_string(None): root}
        self.generate_client_info = options.generate_client_info or generate_client_info
        self.resolver_stats = list()

    async def request_started(self, request, query_string, parsed_query, operation_name, variables, context, request_context):
        self.trace.start_time.GetCurrentTime()
        self.query_string = query_string
        self.document = parsed_query
        self.trace.http.method = self._get_http_method(request)

        client_info = generate_client_info(request)
        if client_info:
            self.trace.client_version = client_info.client_version or ''
            self.trace.client_reference_id = client_info.client_reference_id or ''
            self.trace.client_name = client_info.client_name or ''

        async def on_request_ended(errors):
            start_nanos = self.trace.start_time.ToNanoseconds()
            now = Timestamp()
            now.GetCurrentTime()
            self.trace.duration_ns = now.ToNanoseconds() - start_nanos
            self.trace.end_time.GetCurrentTime()

            op_name = self.operation_name or ''
            self.trace.root.MergeFrom(self.nodes.get(''))
            document = request_context.get('document', None)
            document_ast = document.document_ast if document else None
            await self.add_trace(op_name, document_ast, self.query_string, self.trace)

        return on_request_ended

    async def parsing_started(self, query_string):
        return None

    async def validation_started(self):
        return None

    async def execution_started(self, schema, document, root, context, variables, operation_name, request_context):
        if operation_name:
            self.operation_name = operation_name
        request_context['document'] = document

    async def will_resolve_field(self, root, info, **args):
        if not self.operation_name:
            self.operation_name = '' if not info.operation.name else info.operation.name.value

        node = self._new_node(info.path)
        node.start_time = now_ns() - self.start_time
        node.type = str(info.return_type)
        node.parent_type = str(info.parent_type)

        async def on_end(errors=None, result=None):
            node.end_time = now_ns() - self.start_time

        return on_end

    async def will_send_response(self, response, context):
        root = self.nodes.get('', None)
        root.end_time = now_ns()

        if hasattr(response, 'errors'):
            errors = response.errors
            for error in errors:
                node = root
                if hasattr(error, 'path'):
                    specific_node = self.nodes.get(error.path.join('.'))
                    if specific_node:
                        node = specific_node

                if hasattr(self.options, 'mask_error_details') and self.options.mask_errors_details:
                    error_info = {
                        'message': '<masked>'
                    }
                else:
                    error_info = {
                        'message': str(error),
                        'json': json.dumps(error)
                    }
                node.error.add(error=Trace.Error(**error_info))

    def _get_http_method(self, request):
        try:
            return getattr(Trace.HTTP, request.method.upper())
        except:
            return Trace.HTTP.UNKNOWN

    def _new_node(self, path):
        node = Trace.Node()

        id = path[-1]
        if isinstance(id, Number):
            node.index = id
        else:
            node.response_name = id

        self.nodes[response_path_as_string(path)] = node
        parent_node = self._ensure_parent_node(path)
        n = parent_node.child.add()
        n.MergeFrom(node)
        self.nodes[response_path_as_string(path)] = n
        return n

    def _ensure_parent_node(self, path):
        prev = [''] if len(path) == 1 else path[:-1]
        parent_path = response_path_as_string(prev)
        parent_node = self.nodes.get(parent_path, None)
        if parent_node:
            return parent_node
        return self._new_node(prev)




from __future__ import absolute_import, print_function

import hashlib

from graphene import Schema
from graphql import parse, execute, GraphQLError
from graphql.utils.introspection_query import introspection_query
from json_stable_stringify_python import stringify


def generate_schema_hash(schema):
    # type: (Schema) -> str
    """
    Generates a stable hash of the current schema using an introspection query.
    """
    ast = parse(introspection_query)
    result = execute(schema, ast)

    if result and not result.data:
        raise GraphQLError('Unable to generate server introspection document')

    schema = result.data['__schema']
    # It's important that we perform a deterministic stringification here
    # since, depending on changes in the underlying `graphql-core` execution
    # layer, varying orders of the properties in the introspection
    stringified_schema = stringify(schema).encode('utf-8')
    return hashlib.sha512(stringified_schema).hexdigest()
